
/*
 *  NoNameFS
 *
 *  Copyright (C) 2009 Mike Evans 
 *
 *  NoNameFS has been written to provike discussion on how 
 *  file system naming could be improved so as to be more usable
 *  by human beings.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Authors: Mike Evans <mike@tandem.f9.co.uk>
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <gnome.h>

#include "model.h"

extern  tagfs_state *global_state;

static char *test_data[] = {"line1", "line2", "line3"};


/**
 * Creates a new data model for the MVC pattern 
 * used by the treeviews in the main window
 */
void set_model(GtkWidget *top_widget) {
	
	GtkWidget *treeview;
	GtkListStore *model;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	
	
	// The main one
	treeview = (GtkWidget *)lookup_widget(GTK_WIDGET(top_widget), "treeview1");
	model = gtk_list_store_new  (N_COLUMNS, 
									G_TYPE_INT, // zeroth column holds index
									G_TYPE_STRING, 
									G_TYPE_STRING,
									G_TYPE_STRING,
									G_TYPE_STRING,
									G_TYPE_STRING,
									G_TYPE_STRING,
									G_TYPE_STRING);
	
	
	gtk_tree_view_set_model ( GTK_TREE_VIEW(treeview), GTK_TREE_MODEL(model) );
	
	// the suggested tags one
	treeview = (GtkWidget *) lookup_widget(GTK_WIDGET(top_widget), "treeview2");
	model = gtk_list_store_new  (1, G_TYPE_STRING);
	gtk_tree_view_set_model ( GTK_TREE_VIEW(treeview), GTK_TREE_MODEL(model) );
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Suggestions",
			                                           renderer,
			                                           "text", 0,
			                                           NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
	update_suggested_model(global_state, model);
	
	
}

void tagfs_state_reset(tagfs_state *state) {
	
	state->current_tag_count=0;
	state->possible_tag_count=0;
	state->selected_file_count=0;
	state->heading_tag_count=0;
	state->current_cursor_index=-1;
	
}

tagfs_state *tagfs_state_new() {
	
	tagfs_state *new_state;
	
	new_state = (tagfs_state *) malloc(sizeof(tagfs_state));
	tagfs_state_reset(new_state);
	new_state->database_connection1=NULL;
	new_state->database_connection2=NULL;
	new_state->current_view_columns=0;
	
	return new_state;
	
}

void tagfs_state_free(tagfs_state *old) {

	free(old);
	
}

void goodstrncat( char *to, char *from, int tolimit) {
	
	
	
	tolimit = tolimit - strlen(to) -1;
	if (tolimit > 0 ) strncat(to, from, tolimit);
	
}

void model_get_tags( tagfs_state* state, char *buffer, int buflen ) {
	
	int i;
	*buffer = '\0';
	
	for ( i=0; i < state->current_tag_count; i++) {
		
		goodstrncat(buffer,"/",buflen);
		goodstrncat(buffer,state->current_tags[i].name,buflen);
		if ( state->current_tags[i].data[0] != '\0') {
			goodstrncat(buffer, ":",buflen);
			goodstrncat(buffer, state->current_tags[i].data,buflen);
		}
	}
	
	
	
}

/**
 * get the next whitespace delimited word from the line, limiting the
 * length of the word to max_chars.  Return a pointer the the place
 * we stopped.  line is assummed to be '\0' terminated.
 */
char * parse_word(char *line, char *word, int max_chars) {
	
	max_chars--;  // leave space for termination with '\0'
	
	/* skip leading white space */
	while (*line && isspace(*line)) {
		line++;
	}
	
	/* copy up to max_chars of non-whitespace */
	while (*line && !isspace(*line) && *line!='/' && *line !=':') {
		
		if (max_chars) {
			*word++ = *line;
			max_chars--;
		}

		line++;
		
	}
	
	/* terminate word and return pointer to where we reached */
	*word = '\0';
	return line;
	
}

/**
 * Utility method which returns true (non-zero) if str2 is contained
 * in str1.  The matching is done independent of case.
 */
int str_contains_nocase(char *str1, char *str2) {
	
	char *str2b;
	char *str1b;
	while ( *str1 ) {
		if ( tolower(*str1) == tolower(*str2) ) {
			//start comparison
			str1b = str1;
			str2b = str2;
			while ( *str1b && *str2b && tolower(*str1b)==tolower(*str2b) ) {
				str1b++;
				str2b++;
			}
			// we've stopped.  If we got to the end of str2b then the string was contained
			if ( *str2b=='\0') return 1; //true
		}
		str1++;
		
	}
	// if we got here then str1 was exhausted and no match was found
	return 0; //false
}

/**
 * Utility method to get a given tag from a file if it has it
 * 
 */
tag *file_get_tag ( selected_file *the_file, tag *the_tag ) {
	
	int i;
	
	for ( i=0 ; i < the_file->tag_count ; i++ ) {
		if ( !strcmp(the_file->tags[i].id, the_tag->id)) {
			return &the_file->tags[i];
		}
	}
	return NULL;
}


/**
 * Utility method to check if a file in memory has a given tag
 * 
 */
int file_has_tag ( selected_file *the_file, tag *the_tag ) {
	
	int i;
	
	for ( i=0 ; i < the_file->tag_count ; i++ ) {
		if ( !strcmp(the_file->tags[i].id, the_tag->id)) {
			// ok so it's got the tag - does it have the matching data?
			if ( the_tag->data[0]!='\0') {
				// check each word from the tag's data appears in the filetag's data
				// returns false immediately
				char *data_pointer;
				char word[20];
				data_pointer = the_tag->data;
				while ( *data_pointer ) {
					data_pointer = parse_word(data_pointer, word, sizeof(word));
					if ( !str_contains_nocase(the_file->tags[i].data, word)) {
						return 0; // false
					}
				}				
			}
			/* if we got here either the tag and the data matched or there was
			 * no data and the tag matched so return true
			 */
			
			return 1; //true
		}
	}
	return 0; // false

}

/**
 * utillity method to find out if a tag can have data
 */
int tag_has_datatype(tag *the_tag) {
	if ( the_tag->data_type[0] && strcmp("none",the_tag->data_type)) {
		return 1; //true
	}
	return 0; //false
}

/**
 * First take the selected tags which can have data, then the
 * possible tags that have data
 */
void find_display_tags(tagfs_state *state) {
	
	int i;
	// remove display tags
	state->heading_tag_count = 0;
	// first for current tags
	for ( i=0 ; i < state->current_tag_count ; i++) {
		if ( state->heading_tag_count >= STATE_MAX_HEADING_TAGS) {
			return;  // got enough already
		}
		if ( tag_has_datatype(&state->current_tags[i])) {
			// got one - add this in to heading
			memcpy(&state->heading_tags[state->heading_tag_count],
					&state->current_tags[i], sizeof(tag));
			state->heading_tag_count++;
		}
	}
	// then for possible tags
	for ( i=0 ; i < state->possible_tag_count ; i++) {
		if ( state->heading_tag_count >= STATE_MAX_HEADING_TAGS) {
			return;  // got enough already
		}
		if ( tag_has_datatype(&state->possible_tags[i])) {
			// got one - add this in to heading
			memcpy(&state->heading_tags[state->heading_tag_count],
					&state->possible_tags[i], sizeof(tag));
			state->heading_tag_count++;
		}
	}

		
	
}

void set_view_column(tagfs_state *state, GtkTreeView *treeview, int colno, char *heading) {
	
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	
	if ( colno < state->current_view_columns ) {
		// already have a column - just set the title
		GtkTreeViewColumn *col = gtk_tree_view_get_column(treeview,colno);
		gtk_tree_view_column_set_title(col,heading);
	} else {
		// need to make a column
		renderer = gtk_cell_renderer_text_new ();
		column = gtk_tree_view_column_new_with_attributes (heading,
				                                           renderer,
				                                           "text", colno+1,
				                                           NULL);
		gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
	}
	
}

void update_view(tagfs_state *state, GtkTreeView *treeview) {
	
	GtkTreeViewColumn *column;
	int i, extra;
	
	// create or adjust any existing columns
	for (  i=0 ; i < state->heading_tag_count && i < N_COLUMNS-1 ; i++) {
		set_view_column(state, treeview, i, state->heading_tags[i].name);
	}
	
	// add one column for 'other' tags on the end
	set_view_column(state, treeview, i++, "");
	
	// now remove any surplus ones
	extra = state->current_view_columns - state->heading_tag_count -1 ;
	while (extra > 0 ) {
		column = gtk_tree_view_get_column(treeview, i);
		gtk_tree_view_remove_column(treeview, column);
		extra--;
	}
	
	// finally remember how many we have in the state - which is not an ideal
	// thing to do but getting the count from the view was too painful!
	state->current_view_columns = state->heading_tag_count + 1;
	
	
	
}

void update_suggested_model(tagfs_state *state, GtkListStore *liststore ) {
	
	int i; 
	GtkTreeIter  iter;
	
	gtk_list_store_clear(liststore);
	for ( i=0 ; i<state->possible_tag_count ; i++ ){
		gtk_list_store_append (liststore, &iter);
		gtk_list_store_set (liststore, &iter,
					        0, state->possible_tags[i].name,
					        -1);
	}
}

void update_model(tagfs_state *state, GtkWidget *top_widget) {
	
	GtkTreeView *treeview;
	GtkListStore *liststore;
	GtkTreeIter  iter;
	int i,j;

	
	
	treeview = (GtkTreeView *) lookup_widget(GTK_WIDGET(top_widget), "treeview1");
	update_view(state, treeview);
	
	liststore = GTK_LIST_STORE(gtk_tree_view_get_model(treeview));
		
	// remove all the data
	gtk_list_store_clear(liststore);
	//  populate the data
	for ( i=0; i< state->selected_file_count; i++) {
		// clear the used_header flags
		for (j=0; j < state->selected_files[i].tag_count; j++) {
			state->selected_files[i].tags[j].used_header = FALSE;
		}
		
		// point the iterator to the end of the list
		gtk_list_store_append (liststore, &iter);
		
		// put the index to the file in the zeroth column of the model
		gtk_list_store_set (liststore, &iter, 0, i, -1);
		
		// fill up the columns with the headers
		for ( j=0 ; j < state->heading_tag_count ; j++) {
			char *value;
			tag *the_tag;
			the_tag = file_get_tag(&state->selected_files[i], &state->heading_tags[j]);
			if (the_tag){
				value = the_tag->data;
				the_tag->used_header = TRUE;
			} else {
				value = "";
			}
				
			gtk_list_store_set (liststore, &iter,
			                          j+1, value,
			                          -1);
		}
		
		// make a column with any unused headers in it
		state->selected_files[i].unused_headers[0]='\0';
		int k;
		for ( k=0; k < state->selected_files[i].tag_count; k++) {
			if (state->selected_files[i].tags[k].used_header == FALSE ) {
				goodstrncat(state->selected_files[i].unused_headers,
								            "/", FILE_UNUSED_HEADER_LEN);
				goodstrncat(state->selected_files[i].unused_headers,
				            state->selected_files[i].tags[k].name, FILE_UNUSED_HEADER_LEN);
			}
		}
		
		gtk_list_store_set( liststore, &iter, j+1, state->selected_files[i].unused_headers,-1);
	}
	
	// now update the suggested tags
	treeview = (GtkTreeView *) lookup_widget(GTK_WIDGET(top_widget), "treeview2");
	liststore = GTK_LIST_STORE(gtk_tree_view_get_model(treeview));
	update_suggested_model(state, liststore);

	
	
}



